
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;; Module de documentation de la mmoire en mode protg. ;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; I ;;;;;;;;;;;;;;;;;;;;; Mmoire tendue sur PC 386 ;;;;;;;;;;;;;;;;;;;;;;;
if 0 ; remarques

  Le seul support par le BIOS (la ROM) pour grer la mmoire tendue consiste
en deux fonctions greffes sur l'interruption 15h du BIOS, initialement
conue pour grer les cassettes de l'IBM PC: l'une renvoie la taille de la
mmoire tendue, l'autre permet (avec bien peu d'efficacit) d'effectuer des
copies de mmoire  partir (ou vers) la mmoire tendue.
  Fonction AH=88H: Renvoie dans AX la taille en Ko (jusqu' 64Mo donc) de la
mmoire tendue, c'est  dire  partir de l'adresse physique 100000H. Cette
mthode incite chacun  prendre toute la mmoire tendue existante; mais on
pourra ne prendre que la mmoire ncessaire, en haut de la mmoire tendue,
et dtourner l'interruption 15h, pour renvoyer la taille tronque avec la
fonction 88h (mthode dite INT 15).
  L'ennui est que VDISK ne tient pas compte de ces recommandations, et, sans
les bonnes options disponibles dans les versions rcentes, se place juste
aprs la limite du mga, en y crivant son BPB !!! donc,  exclure.
Nanmoins, un programme soucieux de compatibilit devra tenir compte de
cette ventualit.
  Fonction AH=87H: passe en mode protg pour effectuer un mouvement de
blocs mmoire. ES:SI doit pointer sur la GDT, dont voici le format:
00h -> (dummy) entre nulle
08h -> GDT
10h -> start: segment source
18h -> dest: segment destination
20h -> BIOS CS
28h -> stack
  Au retour, AH contient le code d'erreur (ou 0 si tout s'est droul
correctement): 1 pour une erreur de parit (typiquement de la mmoire
inexistante), 2 pour une mauvaise GDT, 3 pour l'impossibilit d'effectuer
l'opration (sur un 286, ou en mode virtuel).
  Si cette fonction existe sur un AT 286, elle devra rinitialiser le
processeur pour revenir en mode rel. En tout cas, cette fonction est lente,
et peu efficace.

endif

;; II ;;;;;;;;;;;;;;;;;;; Contrle de la ligne A20 ;;;;;;;;;;;;;;;;;;;;;;;;;;
if 0 ; remarques

  Par souci de compatibilit avec le 8086, qui calculait ses adresses
physiques sur 20 bits, et pour lequel une adresse au-dessus de 100000H
tait considre en tronquant le bit 21, donc en revenant au dbut (phnomne
appel wrap-around en anglais, ou enroulement des adresses), les AT, qui eux
calculent leurs adresses physiques sur 24 ou 32 bits, ont t quips d'un
systme qui bloque  0 la ligne d'adresse A20 (la 21me, la premire tant
A0), provoquant ainsi l'enroulement des adresses que certains vieux
programmes utilisent. Tout programme voulant utiliser la mmoire tendue doit
d'abord librer la ligne A20 (au cas ou elle serait bloque).
  Le contrle de la ligne se fait par le port de contrle du clavier.

endif

;;;; Constantes pour accder  la ligne A20
KB_command_port	= 64h		;(en criture)
KB_status_port	= 64h		;(en lecture)
KB_data_port	= 60h
IB_free		= 2		;pour savoir si le port clavier est libre
KB_command_A20	= 0D1h		;pour accder au port de la ligne A20
set_A20_ON  = 11011111b
set_A20_OFF = 11011101b

;;;; Macros pour contrler la ligne A20
Macro	KB_ready	erreur,nombre_essai
   Local wait_IB
 ;; Cette macro teste si le port clavier est disponible (chose ncessaire
 ;; avant de lui envoyer des commandes)
 ;; - erreur doit tre une tiquette ou se brancher si la ligne
 ;; n'est pas disponible.
 ;; - nombre_essai contient le nombre d'essais d'accs au port avant de
 ;; s'avouer vaincu.
 ;; ** dtruit AX,CX
   ifnb <&nombre_essai>
     mov cx,nombre_essai
   else
     mov cx,100
   endif
  wait_IB:
   in al,kb_status_port
   test al,ib_free
   loopne wait_IB
   jcxz erreur
EndM			;;      KB_ready

Macro	set_A20		ON_OFF,erreur,nombre_essai
 ;;  Cette macro active ou dsactive la ligne A20, en ayant test que le port
 ;; clavier est disponible.
 ;; - ON_OFF contient ON ou OFF selon que l'on veuille activer ou dsactiver
 ;; la ligne. Si ce paramtre est absent, AH doit contenir A20_ON ou A20_OFF.
 ;; - erreur doit tre une tiquette ou se brancher si la ligne
 ;; n'est pas disponible.
 ;; - nombre_essai contient le nombre d'essais d'accs au port avant de
 ;; s'avouer vaincu.
 ;; ** dtruit AX
  ifnb <&ON_OFF>
    mov ah,A20_&ON_OFF
  endif
  KB_ready
  mov al,KB_command_A20
  out KB_command_port,al
  mov al,ah
  out KB_command_port,al
EndM

Macro	Test_A20
  Local Exit_
   ;;  Teste si la ligne A20 est active, en vrifiant si oui ou non les
   ;; adresses mmoire s'enroulent aprs la limite du mga-octet.
   ;; * Renvoie C ssi la ligne est active
   ;; ** dtruit AX
   push ds
   push es
   cli
   sub ax,ax
   mov es,ax		; es=$0000
   dec ax
   mov ds,ax		; ds=$FFFF
   mov ax,ds:[-2]	;; vrifier [$10FFEE]=[$0FFEE]
   cmp ax,es:[-18]
   stc
   jnz Exit_
   inc ds:[-2]		;; vrifier [$10FFEE]=[$0FFEE] aprs modification
   cmp ax,es:[-18]
   mov ds:[-2],ax
   stc
   jz Exit_
   clc
  Exit_:
   sti
   pop es
   pop ds
EndM

if 0
Exemple d'utilisation de ces macros

Proc	Teste_A20	near
 ;; Renvoie cf ssi A20 activ
 ;; * dtruit ax
   Test_A20
   ret
EndP
Proc	Controle_A20	near
 ;; met la ligne A20 selon cf (cf=1 => activ)
   mov ax,1110111010000000b	;ax = $dd00 shr 1 + $8000
   adc ah,al			;ah = $dd shr 1 + cf
   adc ah,ah			;ah = $dd + ex-cf*2
   set_A20
   ret
EndP	Controle_A20
Proc	Enable_A20	near
 ;; active s'il le faut la mmoire tendue
   call Teste_A20
   @if nc
     stc
     call controle_A20
   @endif
EndP	Enable_A20

EndIf

;; III ;;;;;;;;;;;;;;;;;;;;;;; Le standard XMS ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
if 0 ; remarques

  Heureusement, un standard s'est impos pour grer la mmoire tendue, le
format XMS. Ce standard gre trois sortes de mmoire:
- la HMA, High Memory Area, que le XMS gre depuis ses premires versions.
Elle commence en 0FFFFh:0010h et se termine en 0FFFFh:0FFFFh, soit en
adresse physique de 100000h  10FFEFh, (de 1024Ko  1088Ko non compris), soit
juste aprs la limite du mgaoctet, et est accessible en mode rel (ou 8086
virtuel) quand la ligne A20 est libre. Des programmes rsidents et autres
gestionnaires du DOS s'y placent. Comme on y accde  partir du segment
0FFFFh, la mmoire s'alloue dans le segment, et un programme s'y plaant doit
prendre la responsabilit d'tre xcutable; de plus la ligne A20 pouvant
tre bloque, il est conseiller d'y placer des donnes, ou de n'en xcuter
des programmes qu'aprs les avoir recopis en mmoire conventionnelle.
- les blocs EMB (extended memory blocks), de mmoire tendue. Ces blocs
sont des blocs de la mmoire situe aprs le mgaoctet, la HMA et/ou les
disques virtuels. Ils sont allous au kilo (octet), et grs par le XMS.
Le XMS peut faire varier leur adresse, pour les regrouper et assigner un
EMB de taille trop grande sinon. Aussi, si l'on veut qu'un bloc reste en
place, il faut le vrouiller. On peut vrouiller plusieurs fois un bloc,
et le XMS retient le nombre de vrouillages jusqu' 255. Ainsi, plusieurs
routines assez indpendantes peuvent utiliser le mme bloc.
- les blocs UMB (upper memory blocks). Ce sont des segments en mmoire
conventionnelle, qui se placent par-dessus les adresses rserves pour la ROM
mais inutilises gnralement, de C000  EFFF. Pour aller par-dessus la ROM,
le XMS utilise la pagination en mode 8086 virtuel (avec EMM386).

* Pour accder au driver XMS, il faut prendre son adresse grce 
l'interruption (dite multiplexeur) int 2Fh. D'abord, vrifier la prsence du
XMS avec AX=4300h. Le driver est prsent si la rponse est AL=80h. Dterminer
alors l'adresse du driver avec AX=4310h. ES:BX contiendra l'adresse d'une
routine far se chargeant du bouleau comme une routine d'interruption, par
fonction dans AH. DX contiendra un paramtre, et le rsultat. AX sera  1 en
cas de russite,  0 en cas d'erreur, et BH contiendra alors le code
d'erreur. En voil la liste:
00h -> version du gestionnaire
  >> Le format des versions est en BCD compact, hi.lo : lire 0301 comme
 version 3.01. AX contient le numro de version, BX celui de rvision. DX
 est nul ou  1,  1 ssi la HMA est prsente.
01h -> allouer HMA
  << DX: taille en octets
  >> DX: offset dans le segment 0FFFFh
02h -> librer HMA
  << DX: offset dans 0FFFFh
03h -> activer A20 globalement
04h -> dsactiver A20 globalement
05h -> activer A20 localement
06h -> dsactiver A20 localement
07h -> tat A20
08h -> taille de la mmoire tendue
  >> DX taille totale, AX taille libre
09h -> Allouer EMB
  << DX: taille demande en Ko
  >> DX: handle  conserver pour accder au bloc
0Ah -> Librer EMB
  << DX: handle
0Bh -> Dplacer bloc mmoire
  << DS:DX pointe sur la structure suivante:
   00 (DWORD) longueur en octets  transfrer
   04 (WORD)  handle source
   06 (DWORD) offset source
   0A (WORD)  handle destination
   0C (DWORD) offset destination
   Le handle 0 correspond  la mmoire conventionnelle, et l'offset sur 32 bits
 est alors remplac par offset 16 bits suivi de numro de segment. Les EMB
 doivent tre vrouills avant d'effectuer le dplacement (ou est-ce seulement
 pour la fonction 0Fh ?).
   Remarque: pour plus de rapidit, aligner les offsets !
0Ch -> Vrouiller EMB
  << DX: handle
  >> BX:DX adresse physique du bloc (DS:DX ?)
0Dh -> Dvrouiller EMB
  << DX: handle
0Eh -> Lire info sur EMB
  << DX: handle
  >> BH: compteur du nombre de vrouillages,
     BL: nombre de handles disponibles
     DX: taille de l'EMB en Ko
0Fh -> Changer taille EMB
  << DX:handle ( vrouiller avant)
     BX:nouvelle taille.
   Si la taille du bloc est rduite, le contenu est perdu.
10h -> Allouer UMB
  << BX:taille en paragraphes
  >> BX:numro de segment
11h -> Librer UMB
  << BX:numro de segment

Codes d'erreur hxadicimaux (dans BH si AX=0)
80	fonction inconnue
81	RAMDISK ou VDISK prsent ! (contre HMA)
82	erreur A20
8E	erreur gnrale
8F	erreur irrcuprable
90	pas de HMA
91	HMA occupe
92	bloc trop petit pour HMA
93	HMA non alloue
94	ligne encore en fonction (contre rentrance du serveur XMS)
A0	plus de XMS disponible
A1	tous handles XMS occups
A2	handle incorrect
A3	handle source mauvais
A4	offset source mauvais
A5	handle destination mauvais
A6	offset destination mauvais
A7	longueur mauvaise
A8	superposition interdite pour un move
A9	erreur de parit
AA	EMB non vrouill
AB	EMB encore vrouill
AC	dbordement compteur vrouillage EMB
AD	EMB ne peut pas tre vrouill
B0	UMB plus petit disponible
B1	plus de UMB disponible
B2	adresse segment UMB incorrecte

endif

Struc	XMS_Move_Params		;; paramtres pour la fonction XMS 0Bh
   LengthToMove	dd	?
   SourceHandle	dw	?
   SourceOffset	dd	?
   DestHandle	dw	?
   DestOffset	dd	?
EndS	XMS_Move_Params



;; IV ;;;;;;;;;;;;; La gestion de la mmoire en mode protg ;;;;;;;;;;;;;;;;
if 0	; commentaires

Rappels sur le codage des DT

Slecteur
bits 15-3 = index dans DT
bit  2    = TI, 1 si LDT, 0 si GDT
bits 1-0  = RPL, privilge du demandeur d'accs au segment

Descripteur
6  bbxl		l: quartets du champ limite
4  ppbb		b: quartets du champ base
2  bbbb		x: quartet G,D,0,AVL
0  llll		pp: octet des droits d'accs, P,DPL,DT,Type

G: Granularit, 0 si limite est en octets, 1 si limite est en 4K
D: Dfaut, 0 si segment (code/pile) 16 bits, 1 si segment (code/pile) 32 bits
AVL: (Available) libre
P: prsent, 1 si prsent (sinon, exception 0Bh, ou 0Ch si pile)
DPL: Descriptor Privilege Level (00:plus fort,11:plus faible), sur 2 bits
DT: Descriptor Type (1:normal)
Type:
. si DT=1,
 0EWA pour les donnes; E:extensible vers le bas; W:modifiable (writable)
 1CRA pour le code    ; C:docile (comforming)   ; R:lisible (readable)
  A:accessible
. si DT=0,
 0000	rserv				1000	rserv
 0001	TSS 16bits disponible		1001	TSS 32bits disponible
 0010	LDT				1010	rserv
 0011	TSS 16bits occupe		1011	TSS 32bits occupe
 0100	Porte d'appel 16bits		1100	Porte d'appel 32bits
 0101	Porte de tche			1101	rserv
 0110	Porte d'interruption 16bits	1110	Porte d'interruption 32bits
 0111	Porte de trappe	16bits		1111	Porte de trappe 32bits

Format d'une porte:
* Porte de Tche:
 dw rserv
 dw slecteur segment TSS
 dw p_dpl_0_type_rserv
 dw rserv
* Porte d'Interruption ou de Trappe:
 dw offset_lo
 dw slecteur segment code
 db 000_rserv,p_dpl_0_type
 dw offset_hi
  La diffrence est qu'un appel  une porte de trappe ncessite CPL<IOPL
* Porte d'Appel:
 dw offset_lo
 dw slecteur segment code
 db 000_compteur,p_dpl_0_type
 dw offset_hi
 Compteur: nombre d'oprandes  recopier dans la pile, en mot (16 ou 32 bits
selon la nature de la porte)

EndIf

Macro	DefDescriptor	bas,Lim,_G_D_0_AVL,_P_DPL_DT_Type
    dw  (Lim) and 0FFFFh
    dw	(bas) and 0FFFFh
    db	( (bas) shr 16 ) and 0FFh
    db	_P_DPL_DT_Type
    db	(_G_D_0_AVL) shl 4 + ( ( (Lim) shr 16) and 0Fh )
    db	(bas) shr 24
EndM

Macro	DefPorte	sel,ofs,_P_DPL_DT_Type,compteur
    dw  (ofs) and 0FFFFh
    dw	sel
    db	compteur
    db	_P_DPL_DT_Type
    dw (ofs) shr 16
EndM

Macro	SetDescriptor	DESCRIPT
   ;;; Cette macro initialise un descripteur avec
   ;;; - DESCRIPT contenant la chane adresse du descripteur
   ;;; - ebx pour adresse de base
   ;;; - ecx pour champ limite
   ;;; - al pour les champs d'information: P,DPL,DT,Type
   ;;; - ah pour G,D,AVL
   ;;; * dtruit eax,ebx,ecx
   mov [DESCRIPT],cx
   shr ecx,16
   mov [DESCRIPT+2],bx
   shr ebx,16
   or cl,ah
   mov ch,bh
   mov bh,al
   mov [DESCRIPT+4],bx
   mov [DESCRIPT+6],cx
EndM

Macro	GetDescriptor	DESCRIPT
   Local
   ;;; Cette macro fait le contraire de la prcdente:
   ;;; - DESCRIPT contient la chane adresse du descripteur
   ;;; - ebx pour adresse de base
   ;;; - ecx pour champ limite
   ;;; - al pour les champs d'information: P,DPL,DT,Type
   ;;; - ah pour G,D,AVL
   mov cx,DESCRIPT[6]
   mov bx,DESCRIPT[4]
   mov al,bh
   mov bh,ch
   mov ah,cl
   and cx,0Fh
   shl ebx,16
   mov bx,DESCRIPT[2]
   shl ecx,16
   mov cx,DESCRIPT
EndM

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; TSS ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Struc	TSS16
;;;; spcifications Intel
    Sel_Lien_Arriere dw ?
    SP_CPL0	dw	? ; non mis  jour
    SS_CPL0	dw	? ;
    SP_CPL1	dw	? ;
    SS_CPL1	dw	? ;
    SP_CPL2	dw	? ;
    SS_CPL2	dw	? ;
    saveIP	dw	?
    saveFlags	dw	?
    saveAX	dw	?
    saveCX	dw	?
    saveDX	dw	?
    saveBX	dw	?
    saveSP	dw	?
    saveBP	dw	?
    saveSI	dw	?
    saveDI	dw	?
    saveES	dw	?
    saveCS	dw	?
    saveSS	dw	?
    saveDS	dw	?
    saveLDT	dw	? ; non mis  jour
ends

Struc	TSS32
;;;; spcifications Intel
    Sel_Lien_Arriere dw ?,0	;(???)
    ESP_CPL0	dd	?	; non mis  jour
    SS_CPL0	dw	?,0	;
    ESP_CPL1	dd	?	;
    SS_CPL1	dw	?,0	;
    ESP_CPL2	dd	?	;
    SS_CPL2	dw	?,0	;
		dd	0	;(rserv)
    saveEIP	dd	?
    saveEFlags	dd	?
    saveEAX	dd	?
    saveECX	dd	?
    saveEDX	dd	?
    saveEBX	dd	?
    saveESP	dd	?
    saveEBP	dd	?
    saveESI	dd	?
    saveEDI	dd	?
    saveES	dw	?,0	; les 0 ne sont pas mis  jour
    saveCS	dw	?,0
    saveSS	dw	?,0
    saveDS	dw	?,0
    saveFS	dw	?,0
    saveGS	dw	?,0
    saveLDT	dw	?,0	; non mis  jour
    TrapBit	dw	0	; non mis  jour
    BaseCarteES	dw	?	; non mis  jour
ends

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; IDT ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
if 0
Rappels interruptions:
** Intel **				IP?	CodeErr	CPU	Note
00	Division par 0		F	O	N	0	[1]
01	Pas--pas/dbogage	T/T,F	N	N	0/3	[2]
02	NMI			T	N	N	0	[3]
03	Point d'arrt		L,T	N	N	0	[4]
04	INTO			L,T	N	N	0	[5]
05	BOUND			L,T	O	N	2
06	Opcode invalide		F	O	N	2
07	Coprocesseur absent	F	O	N	2
08 !	Double faute		A	O	O (=0)	2
09 !	Segment dbord copro	A	N	N	2	[6]
0A !	TSS invalide		F	O	O	3	[7]
0B !	Segment absent		F	O	O	3
0C !	Faute de pile		F	O	O	3
0D !	Protection gnrale	F,A	O	O	2
0E !	Faute de page		F	O	O	3
0F !	Rserv intel		-	-	-	-
10 !	Erreur coprocesseur	F	O	N	2	[8]
11 !	vrification alignement	F	O	O	4
12-1F !	rserv intel

* F: faute, T:trappe, A:arrt, L:logicielle
* IP?: oui ssi CS:(E)IP en pile pointe sur l'instruction fautive
* CodeErr: oui ssi code erreur en pile: on push (sur 32 bits si pile 32 bits)
ce code est un slecteur, o RPL a t remplac par bit 0:EXT, et bit 1:IDT
si IDT=1, l'entre est dans l'IDT; si EXT=1, l'interruption est externe.
* CPU: n CPU  partir duquel l'interruption existe, 0:8086, x:80x86
* Notes:
[1] sur 8086/88, trappe, donc CS:IP sur instruction suivante
[2] pour dbogage, T ou F selon.
[3] NMI seulement par signaux externes
[4] CS:IP pointe sur l'octet aprs le point d'arrt
[5] INTO peut tre excute sans qu'il y ait eu dpassement en forant OF=1
[6] code erreur toujours nul
[7] normalement pas sur 386, remplac par 0Dh sur 486
[8] CS:IP pointe sur l'instruction o l'erreur a t dtecte, non gnre.

** BIOS **
05	Print Screen !!!
08	interruption externe!: horloge
09	interruption externe!: clavier
0B	interruption externe!: COM1
0C	interruption externe!: COM2 (0 pour le bios)
0D	interruption externe!: HardDisk
0E	interruption externe!: FloppyDisk
0F	interruption externe!: PRiNter
10	BIOS Video
11	BIOS dtermination Equipement
12	BIOS dtermination taille mmoire
13	BIOS I/O disques
14	BIOS I/O COM
15	BIOS I/O sur cassette
16	BIOS I clavier
17	BIOS O PRN
18	ROM Basic
19	BootStrap
1A	BIOS Horloge
1B	Ctrl-Break
1C	Contrle du compteur Horloge
1D	Table des param. vido
1E	Table de l'unit disquette
1F	Table de donnes supplmentaires CGA (ascii 80-FF)
20-3F	DOS
40	contrleur de disquette PC/XT
41	Table des param. disque dur
44	Table donnes caractres supplmentaires EGA/VGA (0-FF)

Comme on voit, il y a conflit entre les interruptions externes et internes.
Ou bien il faut rediriger les interruptions externes, ou bien il faut utiliser
le champ EXT du code d'erreur (quand il y en a) pour dterminer si
l'interruption est interne ou externe.

endif


;; Appendice A ;;;;;;;;; Dterminer la nature du processeur ;;;;;;;;;;;;;;;;;
if 0 ; remarques
    Dans cet appendice est propose une macro dfinissant une procdure de
  dtermination du processeur courant. Pour l'utiliser, il suffit d'insrer
  la macro dans le segment de code: < def_proc_TestCPU near >. La procdure
  dfinie s'appelle TestCPU, et ncessite que l'assembleur soit en mode
  386.
    Cette procdure est adapte du livre "Processeur et Coprocesseur"
  de Robert Hummel. Elle pourrait tre amliore pour dterminer le step
  prcis du processeur, et pour diffrencier 287XL de 80287 simple.
endif

Macro	def_proc_TestCPU  distance
NDP_STATUS dw -1
proc	TestCPU distance
  ;;;; Procdure prise  Robert L. Hummel dans son bouquin sur les 80x86
  ;
  ; AH
  ; 01	8088 ou 8086
  ; 02	80286
  ; 03	80386
  ; 04	486
  ;
  ; AL
  ; 00	Pas de coprocesseur
  ; 01	8087
  ; 02	80287
  ; 03	80387
  ; 04	486DX ou 487SX
  ;
  ; Note: 0400 <=> 486SX sans copro.
  ;;;;;;
  ; dtruit tous les registres de calcul, registres copro compris.

	MOV	DX,0100H	; rponse par dfaut
	MOV	[NDP_STATUS],-1 ; non zro
  ;; sur 8088/8086, les bits 12  15 de la valeur de Flags pushe sont mis 
  ;; 1 par PUSHF; c'est pourquoi on va essayer de les effacer.
	PUSHF
	POP	AX
	AND	AH,0FH		;(effacer les bits de poids fort)
	PUSH	AX
	POPF
	PUSHF
	POP	AX
	AND	AH,0F0H		;(les rcuprer)
	CMP	AH,0F0H
	JZ	CPUID_2		; si oui, 8086/8088
  ;; sur le 80286, au contraire, ils sont effacs
	INC	DH
	PUSHF
	POP	AX
	OR	AH,0F0H		;(mettre les bits de poids fort)
	PUSH	AX
	POPF
	PUSHF
	POP	AX
	AND	AH,0F0H		;(les rcuprer)
	JZ	CPUID_2		; si oui, 80286
  ;; sur le 80486, le registre EFLAGS dispose d'un bit d'alignement AC.
	INC	DH
	CLI
	MOV	BP,SP		; sauver SP
	AND	SP,NOT 3	; aligner SP
	PUSHFD
	POP	EAX
	MOV	EBX,EAX		; sauver EFLAGS
	XOR	EAX,00040000H	; complmenter AC
	PUSH	EAX
	POPFD
	PUSHFD
	POP	EAX
	PUSH	EBX
	POPFD			; restaurer AC
	MOV	SP,BP		; restaurer la pile
	XOR	EAX,EBX		; regarder si AC a t modifi
	STI
	JZ	CPUID_2		; c'est un 386
	INC	DH
  ;;;; Maintenant, tester le coprocesseur:
CPUID_2:
	FNINIT				;rinitialiser le mot d'tat
	FNSTSW	[NDP_STATUS]
	CMP	[Byte ptr NDP_STATUS],0	;si non nul, pas de copro
	JNZ	CPUID_3
	FNSTCW	[NDP_STATUS]		;tenter d'crire un mot de contrle
	AND	[NDP_STATUS],103FH
	CMP	[NDP_STATUS],3FH
	JNZ	CPUID_3
  ;;;; le seul cas o le copro peut ne pas correspondre, c'est quand un
  ;;;; 80287 est avec un 80386
	MOV	DL,DH			;initialiser le type de NDP
	CMP	DH,3			;pas 386 => ok
	JNZ	CPUID_3
  ;;; 387/287: le NDP (initialis) doit grer les infinis. Mais le 287 les
  ;;; confond.
	FLD1
	FLDZ
	FDIV
	FLD	ST
	FCHS
	FCOMPP
	FSTSW	[NDP_STATUS]
	MOV	AX,[NDP_STATUS]
	SAHF
	JNZ	CPUID_3			;diffrend:387
	DEC	DL
CPUID_3:
	MOV	AX,DX
	RET
endp	TestCPU
EndM	def_proc_TestCPU

