#ifndef Objects_H
#define Objects_H

#include "MyTypes.h"

#ifdef Objects_G
  #define vextern
#else
  #define vextern extern
#endif

/*OSL preprocessing
about this file:
   Header for "Objects" module
   Defines Objects descriptors, and Class descriptors.

about its implementation:
   Implementation will eventually use Garbage Collecting,
  but no such thing in the header at the time being. Perhaps
  we'll need logon and logoff primitives for object pointers (?).

   Philosophy for this implementation of the OO paradygm:
  - descriptor are needed to link objects one with the other.
  - stable objects are to be compiled not to need run-time help
  anymore. Compile info is provided with objects.

****/

/************************** Objects Types *******************************/

typedef struct TObject {
   Name		     * name ;
   void  	     * body ;   /* the object's data */
   struct TClass     * class ;  /* its class */
   struct TObject    * up ;     /* the object just more global */
} TObject ;


typedef struct TClass {
   Name		     * name ;
   struct TMetaClass * meta ;   /* the class' meta class */
   void              * info ;   /* class info for the meta class */
} TClass ;


typedef char      VInput  (void) ;
typedef void      VOutput (char) ;

typedef void      Creator    (TObject *,...) ;   /* be God for an object */
typedef int	  Surveyor   (TObject *) ;	 /* counts object's length */
typedef void      Destructor (TObject *) ;       /* unlinks object */
typedef TObject * Evaluator  (TObject *,...) ;   /* evaluates an object */
typedef void	  Reader     (TObject *,VInput *) ; /* Reads from input */
typedef void	  Writer     (TObject *,VOutput *) ; /* Writes to output */
typedef void	  Dictionary (TObject *Obj,char*name,TObject*Result) ;

typedef struct TField {
   Name  	  * name ;
   struct TClass  * class ;
   void             (*get) (TObject* Descr,TObject* Object,
			    struct TField* Field) ;
	/* Object, Field -> Descriptor pointing to the field */
   int            offset ;
} TField ;

typedef struct TMetaClass {
   Name           * name ;
   Creator        * New ;
   Evaluator      * eval ;
   Surveyor       * length ;
   Destructor 	  * Destroy ;
   Reader	  * ReadFromBin, * ReadFromText ;
   Writer	  * WriteBinTo,  * WriteTextTo ;
   Dictionary	  * GetMember ;
   int     	    NMembers ;
   TField         * Member ;
   TClass         * InfoClass ;
} TMetaClass ;


/**
 Exemple: metaclass for integers mod n. n will be class info.
**/

/********************* Object handling functions ***********************/

TField * GetMetaMember (TMetaClass *MC,Name *name) ;
void SubMember (TObject*Field,TObject*Obj,Name *name) ;
void * GetMember (TObject*Obj,Name *name) ;
void FindObjectWithMetaClassIn (TObject * Result, char*Name,
				TMetaClass *MC, TObject *Where ) ;

/************************* Predefined Objects **************************/

extern Evaluator Identity ;
extern Surveyor  ConstantLength ;
vextern const TClass     c_Void, c_Char, c_Short, c_Long,
			 c_Object, c_MetaClass, c_Class, c_Field ;
vextern const TMetaClass m_Object, m_MetaClass, m_Class, m_Field,
			 MetaVoid, MetaInt, MetaPointer, MetaArray ;
vextern const TObject    Object, MetaClass, Class, DataField ;



/*********************** Defines ... ***********************************/

#define NEW(a)		((a).class->meta->New(&(a)))
#define DESTROY(a)	((a).class->meta->Destroy(&(a)))
#define EVAL		class->meta->eval
#define SIZEOF(a)	((a).class->meta->length(&(a)))
#define BINREAD(a,I)	((a).class->meta->ReadFromBin(&(a),&(I)))
#define TEXTREAD(a,I)	((a).class->meta->ReadFromText(&(a),&(I)))
#define BINWRITE(a,O)	((a).class->meta->WriteBinTo(&(a),&(O)))
#define TEXTWRITE(a,O)	((a).class->meta->WriteTextTo(&(a),&(O)))
#define MEMBER(a,bc,m)  (((bc*)((a).body))->m)
#define INFO(a,ic,i)	(((ic*)((a).class->info))->i)

#ifdef __MSDOS__
  #define c_Int c_Short
#else
  #define c_Int c_Long
#endif

#undef vextern
#endif /* Objects.h */
